from enum import Enum
from typing import Optional
from uuid import UUID
from pydantic import Field, validator, HttpUrl
from pydantic.generics import GenericModel
from pathlib import Path
import datetime
import re

from .base import BaseMessageComponent, MessageComponentTypes

__all__ = [
    'Plain',
    'Source',
    'At',
    'AtAll',
    'Face',
    'Image',
    'Unknown',
    'Quote',
    'ComponentTypes',
    'SendImage'
]

# original text copied from Tim
qq_emoji_text_list = {
    0:   '[惊讶]',
    1:   '[撇嘴]',
    2:   '[色]',
    3:   '[发呆]',
    4:   '[得意]',
    5:   '[流泪]',
    6:   '[害羞]',
    7:   '[闭嘴]',
    8:   '[睡]',
    9:   '[大哭]',
    10:  '[尴尬]',
    11:  '[发怒]',
    12:  '[调皮]',
    13:  '[呲牙]',
    14:  '[微笑]',
    15:  '[难过]',
    16:  '[酷]',
    17:  '[Empty]',
    18:  '[抓狂]',
    19:  '[吐]',
    20:  '[偷笑]',
    21:  '[可爱]',
    22:  '[白眼]',
    23:  '[傲慢]',
    24:  '[饥饿]',
    25:  '[困]',
    26:  '[惊恐]',
    27:  '[流汗]',
    28:  '[憨笑]',
    29:  '[悠闲]',
    30:  '[奋斗]',
    31:  '[咒骂]',
    32:  '[疑问]',
    33:  '[嘘]',
    34:  '[晕]',
    35:  '[折磨]',
    36:  '[衰]',
    37:  '[骷髅]',
    38:  '[敲打]',
    39:  '[再见]',
    40:  '[Empty]',
    41:  '[发抖]',
    42:  '[爱情]',
    43:  '[跳跳]',
    44:  '[Empty]',
    45:  '[Empty]',
    46:  '[猪头]',
    47:  '[Empty]',
    48:  '[Empty]',
    49:  '[拥抱]',
    50:  '[Empty]',
    51:  '[Empty]',
    52:  '[Empty]',
    53:  '[蛋糕]',
    54:  '[闪电]',
    55:  '[炸弹]',
    56:  '[刀]',
    57:  '[足球]',
    58:  '[Empty]',
    59:  '[便便]',
    60:  '[咖啡]',
    61:  '[饭]',
    62:  '[Empty]',
    63:  '[玫瑰]',
    64:  '[凋谢]',
    65:  '[Empty]',
    66:  '[爱心]',
    67:  '[心碎]',
    68:  '[Empty]',
    69:  '[礼物]',
    70:  '[Empty]',
    71:  '[Empty]',
    72:  '[Empty]',
    73:  '[Empty]',
    74:  '[太阳]',
    75:  '[月亮]',
    76:  '[赞]',
    77:  '[踩]',
    78:  '[握手]',
    79:  '[胜利]',
    80:  '[Empty]',
    81:  '[Empty]',
    82:  '[Empty]',
    83:  '[Empty]',
    84:  '[Empty]',
    85:  '[飞吻]',
    86:  '[怄火]',
    87:  '[Empty]',
    88:  '[Empty]',
    89:  '[西瓜]',
    90:  '[Empty]',
    91:  '[Empty]',
    92:  '[Empty]',
    93:  '[Empty]',
    94:  '[Empty]',
    95:  '[Empty]',
    96:  '[冷汗]',
    97:  '[擦汗]',
    98:  '[抠鼻]',
    99:  '[鼓掌]',
    100: '[糗大了]',
    101: '[坏笑]',
    102: '[左哼哼]',
    103: '[右哼哼]',
    104: '[哈欠]',
    105: '[鄙视]',
    106: '[委屈]',
    107: '[快哭了]',
    108: '[阴险]',
    109: '[亲亲]',
    110: '[吓]',
    111: '[可怜]',
    112: '[菜刀]',
    113: '[啤酒]',
    114: '[篮球]',
    115: '[乒乓]',
    116: '[示爱]',
    117: '[瓢虫]',
    118: '[抱拳]',
    119: '[勾引]',
    120: '[拳头]',
    121: '[差劲]',
    122: '[爱你]',
    123: '[NO]',
    124: '[OK]',
    125: '[转圈]',
    126: '[磕头]',
    127: '[回头]',
    128: '[跳绳]',
    129: '[挥手]',
    130: '[激动]',
    131: '[街舞]',
    132: '[献吻]',
    133: '[左太极]',
    134: '[右太极]',
    135: '[Empty]',
    136: '[双喜]',
    137: '[鞭炮]',
    138: '[灯笼]',
    139: '[发财]',
    140: '[K歌]',
    141: '[购物]',
    142: '[邮件]',
    143: '[帅]',
    144: '[喝彩]',
    145: '[祈祷]',
    146: '[爆筋]',
    147: '[棒棒糖]',
    148: '[喝奶]',
    149: '[下面]',
    150: '[香蕉]',
    151: '[飞机]',
    152: '[开车]',
    153: '[高铁左车头]',
    154: '[车厢]',
    155: '[高铁右车头]',
    156: '[多云]',
    157: '[下雨]',
    158: '[钞票]',
    159: '[熊猫]',
    160: '[灯泡]',
    161: '[风车]',
    162: '[闹钟]',
    163: '[打伞]',
    164: '[彩球]',
    165: '[钻戒]',
    166: '[沙发]',
    167: '[纸巾]',
    168: '[药]',
    169: '[手枪]',
    170: '[青蛙]',
    171: '[茶]',
    172: '[眨眼睛]',
    173: '[泪奔]',
    174: '[无奈]',
    175: '[卖萌]',
    176: '[小纠结]',
    177: '[喷血]',
    178: '[斜眼笑]',
    179: '[doge]',
    180: '[惊喜]',
    181: '[骚扰]',
    182: '[笑哭]',
    183: '[我最美]',
    184: '[河蟹]',
    185: '[羊驼]',
    186: '[Empty]',
    187: '[幽灵]',
    188: '[蛋]',
    189: '[Empty]',
    190: '[菊花]',
    191: '[Empty]',
    192: '[红包]',
    193: '[大笑]',
    194: '[不开心]',
    195: '[Empty]',
    196: '[Empty]',
    197: '[冷漠]',
    198: '[呃]',
    199: '[好棒]',
    200: '[拜托]',
    201: '[点赞]',
    202: '[无聊]',
    203: '[托脸]',
    204: '[吃]',
    205: '[送花]',
    206: '[害怕]',
    207: '[花痴]',
    208: '[小样儿]',
    209: '[Empty]',
    210: '[飙泪]',
    211: '[我不看]',
    212: '[托腮]',
    213: '[Empty]',
    214: '[啵啵]',
    215: '[糊脸]',
    216: '[拍头]',
    217: '[扯一扯]',
    218: '[舔一舔]',
    219: '[蹭一蹭]',
    220: '[拽炸天]',
    221: '[顶呱呱]',
    222: '[抱抱]',
    223: '[暴击]',
    224: '[开枪]',
    225: '[撩一撩]',
    226: '[拍桌]',
    227: '[拍手]',
    228: '[恭喜]',
    229: '[干杯]',
    230: '[嘲讽]',
    231: '[哼]',
    232: '[佛系]',
    233: '[掐一掐]',
    234: '[惊呆]',
    235: '[颤抖]',
    236: '[啃头]',
    237: '[偷看]',
    238: '[扇脸]',
    239: '[原谅]',
    240: '[喷脸]',
    241: '[生日快乐]',
    242: '[Empty]',
    243: '[Empty]',
    244: '[Empty]',
    245: '[Empty]',
    246: '[Empty]',
    247: '[Empty]',
    248: '[Empty]',
    249: '[Empty]',
    250: '[Empty]',
    251: '[Empty]',
    252: '[Empty]',
    253: '[Empty]',
    254: '[Empty]',
    255: '[Empty]',
}


image_regex = {
    'group':  r'(?<=\{)([0-9A-Z]{8})\-([0-9A-Z]{4})-([0-9A-Z]{4})-([0-9A-Z]{4})-([0-9A-Z]{12})(?=\}\..*?)',
    'friend': r'(?<=/)([0-9a-z]{8})\-([0-9a-z]{4})-([0-9a-z]{4})-([0-9a-z]{4})-([0-9a-z]{12})'
}


def get_matched_string(regex_result):
    if regex_result:
        return regex_result.string[slice(*regex_result.span())]


class Plain(BaseMessageComponent):
    type = MessageComponentTypes.Plain
    text: str

    def __str__(self):
        return self.text

    def __repr__(self):
        return f'[Plain: {self.text}]'


class Source(BaseMessageComponent):
    type = MessageComponentTypes.Source
    id: int
    time: datetime.datetime

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Source: id={self.id}, time={self.time}]'
        # return f'[Source: id={self.id}]'


class Quote(BaseMessageComponent):
    type = MessageComponentTypes.Quote
    id: int

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Quote: id={self.id}]'


class At(GenericModel, BaseMessageComponent):
    type = MessageComponentTypes.At
    target: int
    display: str

    def __str__(self):
        return self.display

    def __repr__(self):
        return f'[At: target={self.target}, display={self.display}]'


class AtAll(BaseMessageComponent):
    type = MessageComponentTypes.AtAll

    def __str__(self):
        return '@All'

    def __repr__(self):
        return f'[AtAll]'


class Face(BaseMessageComponent):
    type = MessageComponentTypes.Face
    faceId: int

    def __str__(self):
        return qq_emoji_text_list[self.faceId]

    def __repr__(self):
        return f'[Face: id={self.faceId}, {qq_emoji_text_list[self.faceId]}]'


class Image(BaseMessageComponent):
    type = MessageComponentTypes.Image
    imageId: UUID
    url: Optional[HttpUrl] = None

    @validator('imageId', always=True, pre=True)
    @classmethod
    def imageId_formater(cls, v):
        if isinstance(v, str):
            image_type = 'group'
            uuid_string = get_matched_string(re.search(image_regex[image_type], v))
            if not uuid_string:
                image_type = 'friend'
                uuid_string = get_matched_string(re.search(image_regex[image_type], v))
            if uuid_string:
                return UUID(uuid_string)
        elif isinstance(v, UUID):
            return v

    def __str__(self):
        return ''

    def __repr__(self):
        return f'[Image: {self.imageId}]'

    def as_group_image(self) -> str:
        return f'{{{str(self.imageId).upper()}}}.jpg'

    def as_friend_image(self) -> str:
        return f'/{str(self.imageId)}'


class SendImage:
    """
    class for outbound image from local disk
    """
    path: Path
    uuid: UUID = ''

    def __init__(self, path=None, uuid=None):
        if isinstance(path, str):
            self.path = Path(path)
        elif isinstance(path, Path):
            self.path = path

        if isinstance(uuid, str):
            self.uuid = UUID(uuid)
        elif isinstance(uuid, UUID):
            self.uuid = uuid


class Unknown(BaseMessageComponent):
    type = MessageComponentTypes.Unknown
    text: str

    def __str__(self):
        return ''

    def __repr__(self):
        return '[Unknown]'


class ComponentTypes(Enum):
    Plain = Plain
    Source = Source
    At = At
    AtAll = AtAll
    Face = Face
    Image = Image
    Quote = Quote
    Unknown = Unknown


MessageComponents = {
    'At':      At,
    'AtAll':   AtAll,
    'Face':    Face,
    'Plain':   Plain,
    'Image':   Image,
    'Source':  Source,
    'Quote':   Quote,
    'Unknown': Unknown
}



from typing import List, Dict, Optional
from pydantic import BaseModel

from .base import BaseMessageComponent
from .components import Source


class MessageChain(BaseModel):
    # stores the actual components
    __root__: List[BaseMessageComponent] = []

    def __add__(self, value):
        # merge two message chain or append one component
        if isinstance(value, BaseMessageComponent):
            self.__root__.append(value)
            return self
        elif isinstance(value, MessageChain):
            self.__root__ += value.__root__
            return self

    def __str__(self) -> str:
        return ''.join([str(i) for i in self.__root__])

    @classmethod
    def custom_parse(cls, value: List[Dict]) -> 'MessageChain':
        """
        construct message chain from dict
        :param value: dict contains message components
        :return: MessageChain
        """
        from .components import ComponentTypes
        for i in value:
            if not isinstance(i, dict):
                raise TypeError("invaild value")
        return cls(__root__=[ComponentTypes.__members__[m['type']].value(**m) for m in value])

    def __iter__(self):
        yield from self.__root__

    def __getitem__(self, index):
        return self.__root__[index]

    def has(self, component_class) -> bool:
        """
        test if any item in MessageChain is component_class
        :param component_class: the class for the component
        :return: boolean
        """
        for i in self:
            if isinstance(i, component_class):
                return True
        else:
            return False

    def __len__(self) -> int:
        return len(self.__root__)

    def get_first(self, component_class) -> Optional[BaseMessageComponent]:
        """
        Get the first component with component_class
        :param component_class: the class for the component
        :return: None or the component
        """
        for i in self.__root__:
            if isinstance(i, component_class):
                return i

    def get_all(self, component_class) -> List[BaseMessageComponent]:
        return [i for i in self.__root__ if isinstance(i, component_class)]

    def get_source(self) -> Source:
        result = self.get_first(Source)
        assert isinstance(result, Source)
        return result



from typing import Dict
import httpx
from .log import create_logger
from io import BytesIO

from .exceptions import AuthenticationException, NetworkException,\
    UnknownTargetException, PrivilegeException, BadRequestException, MiraiException


class HttpXClient:
    """HttpClient implemented by httpx."""

    DEFAULT_TIMEOUT = 30

    @staticmethod
    def _check_response(result: httpx.Response, url, method) -> Dict:
        if result.status_code != 200:
            raise NetworkException(f'{url} {method} failed')
        result = result.json()
        if method == 'post':
            status_code = result.get('code')
            if status_code is None:
                raise NetworkException('Empty response')
            if status_code == 0:  # normal
                return result
            elif status_code == 1:
                raise AuthenticationException('Incorrect authKey')
            elif status_code == 2:
                raise AuthenticationException('Bot does not exist')
            elif status_code == 3:
                raise AuthenticationException('Session expired')
            elif status_code == 4:
                raise AuthenticationException('Session is not verified')
            elif status_code == 5:
                raise UnknownTargetException('Message target does not exist')
            elif status_code == 10:
                raise PrivilegeException('Bot does not have corresponding privilege')
            elif status_code == 400:
                raise BadRequestException('Bad Request, please check arguments/url')
            else:
                raise MiraiException('HTTP API updated, please upgrade python-mirai-core')
        elif method == 'get':
            return result

    def __init__(self, base_url: str, timeout=DEFAULT_TIMEOUT):
        self.base_url = base_url
        self.session = httpx.AsyncClient(timeout=timeout)
        self.timeout = timeout
        self.logger = create_logger('Network')

    async def get(self, url, headers=None, params=None, timeout=None):
        if timeout is None:
            timeout = self.timeout
        if url != '/fetchMessage':
            self.logger.debug(f'get {url} with params: {str(params)}')

        response = await self.session.get(self.base_url + url, headers=headers, params=params, timeout=timeout)
        return HttpXClient._check_response(response, url, 'get')

    async def post(self, url, headers=None, data=None, timeout=None):
        if timeout is None:
            timeout = self.timeout

        self.logger.debug(f'post {url} with data: {str(data)}')

        response = await self.session.post(self.base_url + url, headers=headers, json=data, timeout=timeout)
        return HttpXClient._check_response(response, url, 'post')

    async def upload(self, url, headers=None, data=None, file: str = None, timeout=None):
        files = {
            'img': BytesIO(open(str(file.absolute()), 'rb').read())
        }
        self.logger.debug(f'upload {url} with file: {file}')
        response = await self.session.post(self.base_url + url, data=data,
                                           headers=headers, files=files, timeout=timeout)
        return response.text

    async def close(self):
        await self.session.aclose()



class MiraiException(Exception):
    pass


class NetworkException(MiraiException):
    pass


class AuthenticationException(MiraiException):
    pass


class PrivilegeException(MiraiException):
    pass


class UnknownTargetException(MiraiException):
    pass


class BadRequestException(MiraiException):
    pass



import re
from typing import Union, List, Type
from datetime import timedelta
from pathlib import Path
from uuid import UUID
import json
from .log import create_logger
from functools import partial
import asyncio

from .models.message import BotMessage, ImageType, MessageChain,\
    Source, Image, Quote, Plain, SendImage, BaseMessageComponent
from .models.events import *
from .models.entity import Friend, Group, GroupSetting, Member, MemberChangeableSetting
from .network import HttpXClient
from .exceptions import AuthenticationException, MiraiException


ImageRegex = {
    "group": r"(?<=\{)([0-9A-Z]{8})\-([0-9A-Z]{4})-([0-9A-Z]{4})-([0-9A-Z]{4})-([0-9A-Z]{12})(?=\}\..*?)",
    "friend": r"(?<=/)([0-9a-z]{8})\-([0-9a-z]{4})-([0-9a-z]{4})-([0-9a-z]{4})-([0-9a-z]{12})"
}


def get_matched_string(regex_result):
    if regex_result:
        return regex_result.string[slice(*regex_result.span())]


class Events(Enum):
    BotOnlineEvent = BotOnlineEvent
    BotOfflineEventActive = BotOfflineEventActive
    BotOfflineEventForce = BotOfflineEventForce
    BotOfflineEventDropped = BotOfflineEventDropped
    BotReloginEvent = BotReloginEvent
    BotGroupPermissionChangeEvent = BotGroupPermissionChangeEvent
    BotMuteEvent = BotMuteEvent
    BotUnmuteEvent = BotUnmuteEvent
    BotJoinGroupEvent = BotJoinGroupEvent

    GroupNameChangeEvent = GroupNameChangeEvent
    GroupEntranceAnnouncementChangeEvent = GroupEntranceAnnouncementChangeEvent
    GroupMuteAllEvent = GroupMuteAllEvent

    # 群设置被修改事件
    GroupAllowAnonymousChatEvent = GroupAllowAnonymousChatEvent  # 群设置 是否允许匿名聊天 被修改
    GroupAllowConfessTalkEvent = GroupAllowConfessTalkEvent  # 坦白说
    GroupAllowMemberInviteEvent = GroupAllowMemberInviteEvent  # 邀请进群

    # 群事件(被 Bot 监听到的, 为被动事件, 其中 Bot 身份为第三方.)
    MemberJoinEvent = MemberJoinEvent
    MemberLeaveEventKick = MemberLeaveEventKick
    MemberLeaveEventQuit = MemberLeaveEventQuit
    MemberCardChangeEvent = MemberCardChangeEvent
    MemberSpecialTitleChangeEvent = MemberSpecialTitleChangeEvent
    MemberPermissionChangeEvent = MemberPermissionChangeEvent
    MemberMuteEvent = MemberMuteEvent
    MemberUnmuteEvent = MemberUnmuteEvent

    FriendMessage = FriendMessage
    GroupMessage = GroupMessage


def _require_session_key(func):
    async def checker(self: 'Bot', *args, **kwargs):
        if not self.session_key:
            raise AuthenticationException('Session key is not set')
        return await func(self, *args, **kwargs)
    return checker


class Bot:
    """
    See https://github.com/mamoe/mirai-api-http for details
    """
    def __init__(self, qq: int, host: str = '127.0.0.1', port: int = 8080, auth_key: str = 'abcdefgh'):
        self.qq = qq
        self.auth_key = auth_key
        self.base_url = f'http://{host}:{port}'
        self.session = HttpXClient(self.base_url)
        self.session_key = ''
        self.logger = create_logger('Bot')

    async def handshake(self):
        await self.auth()
        await self.verify()

    async def auth(self) -> None:
        """
        post auth key, and get session key
        :return: None
        """
        result = await self.session.post('/auth', data={'authKey': self.auth_key})
        self.session_key = result.get('session')

    @_require_session_key
    async def verify(self) -> None:
        await self.session.post('/verify',
                                data={
                                    'sessionKey': self.session_key,
                                    'qq':         self.qq
                                })

    @_require_session_key
    async def release(self) -> None:
        await self.session.post('/release',
                                data={
                                    'sessionKey': self.session_key,
                                    'qq':         self.qq
                                })

    @staticmethod
    def _handle_target_as(target: Union[Group, Friend, Member, int], as_type: Union[Type[Group], Type[Friend], Type[Member]]):
        """
        convert target to id
        :param target: Union[Group, Friend, Member, int]
        :param as_type: Group, Friend or Member
        :return: id, int
        """
        if isinstance(target, int):
            return target
        elif isinstance(target, as_type):
            return target.id
        else:
            raise ValueError(f'Invalid target as {type(as_type)} object.')

    @_require_session_key
    async def send_friend_message(self,
                                  friend: Union[Friend, int],
                                  message: Union[
                                      MessageChain,
                                      BaseMessageComponent,
                                      List[BaseMessageComponent],
                                      str
                                  ],
                                  quote_source: Union[int, Source] = None) -> BotMessage:
        data = {
            'sessionKey':   self.session_key,
            'target':       Bot._handle_target_as(friend, Friend),
            'messageChain': await self._handle_message_chain(message, Friend)
        }
        if quote_source:
            if isinstance(quote_source, int):
                data['quote'] = quote_source
            elif isinstance(quote_source, Source):
                data['quote'] = quote_source.id
        result = await self.session.post('/sendFriendMessage',
                                         data=data)

        return BotMessage.parse_obj(result)

    @_require_session_key
    async def send_group_message(self,
                                 group: Union[Group, int],
                                 message: Union[
                                     MessageChain,
                                     BaseMessageComponent,
                                     List[BaseMessageComponent],
                                     str
                                 ],
                                 quote_source: Union[int, Source] = None) -> BotMessage:
        data = {
            'sessionKey':   self.session_key,
            'target':       Bot._handle_target_as(group, Group),
            'messageChain': await self._handle_message_chain(message, Group)
        }
        if quote_source:
            if isinstance(quote_source, int):
                data['quote'] = quote_source
            elif isinstance(quote_source, Source):
                data['quote'] = quote_source.id
        result = await self.session.post('/sendGroupMessage',
                                         data=data)
        return BotMessage.parse_obj(result)

    @_require_session_key
    async def recall(self, source: Union[Source, int]) -> None:
        data = {
            'sessionKey':   self.session_key,
        }
        if isinstance(source, int):
            data['target'] = source
        elif isinstance(source, Source):
            data['target'] = source.id
        else:
            raise MiraiException('Invalid source argument')

        await self.session.post('/recall', data=data)

    @property
    @_require_session_key
    async def groups(self) -> List[Group]:
        params = {
            'sessionKey': self.session_key,
        }
        result = await self.session.get('/groupList', params=params)
        return [Group.parse_obj(group_info) for group_info in result]

    @property
    @_require_session_key
    async def friends(self) -> List[Friend]:
        params = {
            'sessionKey': self.session_key,
        }
        result = await self.session.get('/friendList', params=params)
        return [Friend.parse_obj(friend_info) for friend_info in result]

    @_require_session_key
    async def get_members(self, target: Union[Group, int]) -> List[Member]:
        if isinstance(target, int):
            group = target
        else:
            group = target.id
        params = {
            'sessionKey': self.session_key,
            'target': group
        }
        result = await self.session.get('/memberList', params=params)
        return [Member.parse_obj(member_info) for member_info in result]

    @_require_session_key
    async def upload_image(self, image_type: ImageType, image_path: Union[Path, str]) -> Optional[Image]:
        if isinstance(image_path, str):
            image_path = Path(image_path)

        if not image_path.exists():
            raise FileNotFoundError('Image not found.')

        data = {
            'sessionKey': self.session_key,
            'type':       image_type.value
        }
        result = await self.session.upload('/uploadImage', file=image_path, data=data)
        regex = ImageRegex[image_type.value]
        uuid_string = re.search(regex, result)
        if uuid_string:
            return Image(imageId=UUID(get_matched_string(uuid_string)))

    @_require_session_key
    async def fetch_message(self, count: int) -> List[Event]:
        params = {
            'sessionKey': self.session_key,
            'count':      count
        }
        result = await self.session.get('/fetchMessage', params=params)

        for index in range(len(result)):
            if hasattr(Events, result[index]['type']):  # if Event
                if 'messageChain' in result[index]:  # construct message chain
                    result[index]['messageChain'] = MessageChain.custom_parse(result[index]['messageChain'])
                result[index] = Events[result[index]['type']].value.parse_obj(result[index])
        return result

    @_require_session_key
    async def message_from_id(self, source_id: Union[Source, Quote, int]):
        if isinstance(source_id, Source):
            source_id = source_id.id
        elif isinstance(source_id, Quote):
            source_id = source_id.id

        params = {
            'sessionKey': self.session_key,
            'id':         source_id
        }

        result = await self.session.get('/messageFromId', params=params)
        if result.get('type') in (EventTypes.GroupMessageEvent.value, EventTypes.FriendMessageEvent.value):
            if "messageChain" in result:
                result['messageChain'] = MessageChain.custom_parse(result['messageChain'])

            if result.get('type') == EventTypes.GroupMessageEvent.value:
                return GroupMessage.parse_obj(result)
            else:
                return FriendMessage.parse_obj(result)
        else:
            raise TypeError(f'Unknown message type')

    @_require_session_key
    async def mute_all(self, group: Union[Group, int]):
        params = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group)
        }

        await self.session.get('/muteAll', params=params)

    @_require_session_key
    async def unmute_all(self, group: Union[Group, int]):
        params = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group)
        }

        await self.session.get('/unmuteAll', params=params)

    @_require_session_key
    async def get_member_info(self, group: Union[Group, int], member: Union[Member, int]):
        params = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'memberId':   self._handle_target_as(target=member, as_type=Member)
        }

        result = await self.session.get('/memberInfo', params=params)
        return MemberChangeableSetting.parse_obj(result)

    @_require_session_key
    async def get_bot_member_info(self, group: Union[Group, int]):
        return await self.get_member_info(group, self.qq)

    @_require_session_key
    async def set_member_info(self, group: Union[Group, int],
                              member: Union[Member, int],
                              setting: MemberChangeableSetting):
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'memberId':   self._handle_target_as(target=member, as_type=Member),
            'info':       json.loads(setting.json())
        }

        await self.session.post('/memberInfo', data=data)

    @_require_session_key
    async def get_group_config(self, group: Union[Group, int]) -> GroupSetting:
        params = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
        }
        result = await self.session.get('/groupConfig', params=params)
        return GroupSetting.parse_obj(result)

    @_require_session_key
    async def set_group_config(self, group: Union[Group, int],
                               config: GroupSetting):
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'config':     json.loads(config.json())
        }

        await self.session.post('/groupConfig', data=data)

    @_require_session_key
    async def mute(self, group: Union[Group, int],
                   member: Union[Member, int],
                   time: Union[timedelta, int]):
        if isinstance(time, timedelta):
            time = int(time.total_seconds())
        time = min(86400 * 30, max(60, time))  # time should between 1 minutes and 30 days
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'MemberId':   Bot._handle_target_as(target=member, as_type=Member),
            'time':       time
        }
        await self.session.post('/mute', data=data)

    @_require_session_key
    async def unmute(self, group: Union[Group, int],
                     member: Union[Member, int]):
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'MemberId':   Bot._handle_target_as(target=member, as_type=Member)
        }
        await self.session.post('/unmute', data=data)

    @_require_session_key
    async def unmute(self, group: Union[Group, int],
                     member: Union[Member, int],
                     message: str = ''):
        data = {
            'sessionKey': self.session_key,
            'target':     Bot._handle_target_as(target=group, as_type=Group),
            'MemberId':   Bot._handle_target_as(target=member, as_type=Member)
        }
        if message:
            data['msg'] = message

        await self.session.post('/kick', data=data)

    async def _handle_image(self, message: BaseMessageComponent, image_type: ImageType):
        if not isinstance(message, SendImage):
            return json.loads(message.json())

        if message.uuid:
            return {
                'type':    'Image',
                'imageId': message.uuid
            }
        else:
            return {
                'type':    'Image',
                'imageId': await self.upload_image(image_type, message.path)
            }

    async def _handle_message_chain(self, message: Union[
        MessageChain,
        BaseMessageComponent,
        List[BaseMessageComponent],
        str
    ], as_type: Union[Type[Group], Type[Friend]]) -> List:
        if isinstance(message, MessageChain):
            return json.loads(message.json())
        elif isinstance(message, str):
            return [json.loads(Plain(text=message).json())]
        elif isinstance(message, BaseMessageComponent):
            if as_type == Group:
                image_type = ImageType.Group
            else:
                image_type = ImageType.Friend
            return [await self._handle_image(message, image_type)]
        elif isinstance(message, (tuple, list)):
            if as_type == Group:
                image_type = ImageType.Group
            else:
                image_type = ImageType.Friend
            result = [await self._handle_image(m, image_type=image_type) for m in message]
            return result
        else:
            raise ValueError('Invalid message')



import sys

from logging import DEBUG, Formatter, getLogger, Logger, StreamHandler


logger = getLogger('Mirai-core')


def create_logger(module: str) -> Logger:
    """
    create logger for module
    """
    return logger.getChild(module)


def install_logger():
    default_handler = StreamHandler(sys.stderr)
    default_handler.setFormatter(Formatter('[%(asctime)s] %(levelname)s in %(module)s: %(message)s'))
    logger.setLevel(DEBUG)
    logger.addHandler(default_handler)



import asyncio
from typing import DefaultDict, Union, List, Callable, Any, Awaitable
from collections import defaultdict
from dataclasses import dataclass
import signal
from .log import create_logger, install_logger

from .bot import Bot
from .models.events import Event, EventTypes


@dataclass
class EventHandler:
    func: Callable


class Shutdown(Exception):
    pass


class Updater:
    def __init__(self, bot: Bot, loop: asyncio.AbstractEventLoop = None):
        self.bot = bot
        if loop:
            self.loop = loop
        else:
            self.loop = asyncio.new_event_loop()
        self.logger = create_logger('Updater')
        self.event_handlers: DefaultDict[EventTypes, List[EventHandler]] = defaultdict(lambda: list())

    async def run_task(self, shutdown_trigger=None):
        """
        return awaitable coroutine to run in any event loop
        :param shutdown_trigger: shutdown event if running in main thread
        """
        self.logger.debug('Run tasks')
        tasks = [
            self.bot.handshake(),
            self.message_polling(),
        ]
        if shutdown_trigger:
            tasks.append(self.raise_shutdown(shutdown_trigger))
        await asyncio.wait(tasks)

    def add_handler(self, event: Union[EventTypes, List[EventTypes]]):
        """
        Decorator for event listeners
        Catch all is not supported at this time
        :param event: events.Events
        """
        def receiver_wrapper(func):
            if not asyncio.iscoroutinefunction(func):
                raise TypeError("event body must be a coroutine function.")

            # save function and its parameter types
            event_handler = EventHandler(func)
            if isinstance(event, EventTypes):
                # add listener
                self.event_handlers[event].append(event_handler)
            else:
                for e in event:
                    if isinstance(e, EventTypes):
                        self.event_handlers[e].append(event_handler)
            return func

        return receiver_wrapper

    async def message_polling(self, count=5, interval=0.5):
        """
        polling message and fire events
        :param count: maximum message count for each polling
        :param interval: minimum interval between two polling
        """
        while True:
            await asyncio.sleep(interval)

            results: List[Event] = await self.bot.fetch_message(count)
            if len(results) > 0:
                self.logger.debug('Received messages:\n' + '\n'.join([str(result) for result in results]))
            for result in results:
                asyncio.run_coroutine_threadsafe(self.event_caller(result), self.loop)

    async def event_caller(self, event: Event):
        for handler in self.event_handlers[event.type]:
            if await handler.func(event):  # if the function returns True, stop calling next event
                break

    def run(self, log_to_stderr=True):
        asyncio.set_event_loop(self.loop)

        shutdown_event = asyncio.Event()

        def _signal_handler(*_: Any) -> None:
            shutdown_event.set()

        try:
            self.loop.add_signal_handler(signal.SIGTERM, _signal_handler)
            self.loop.add_signal_handler(signal.SIGINT, _signal_handler)
        except (AttributeError, NotImplementedError):
            pass

        if log_to_stderr:
            install_logger()

        self.loop.create_task(self.run_task(shutdown_trigger=shutdown_event.wait))
        self.loop.run_forever()

    async def raise_shutdown(self, shutdown_event: Callable[..., Awaitable[None]]) -> None:
        await shutdown_event()
        await self.bot.release()
        raise Shutdown()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="python-mirai-core",
    version='0.2',
    description='A framework for OICQ(QQ, made by Tencent) headless client "Mirai".',
    author='Chenwe-i-lin, jqqqqqqqqqq',
    author_email="Chenwe_i_lin@outlook.com",
    url="https://github.com/NatriumLab/python-mirai-core",
    packages=find_packages(include=("mirai_core", "mirai_core.*")),
    python_requires='>=3.7',
    keywords=["oicq qq qqbot", ],
    install_requires=[
        "httpx",
        "pydantic"
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: User Interfaces',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        "Operating System :: OS Independent"
    ]
)



from enum import Enum
from pydantic import BaseModel

__all__ = [
    "MessageComponentTypes",
    "BaseMessageComponent"
]


class MessageComponentTypes(Enum):
    Source = "Source"
    Plain = "Plain"
    Face = "Face"
    At = "At"
    AtAll = "AtAll"
    Image = "Image"
    Quote = "Quote"
    Unknown = "Unknown"


class BaseMessageComponent(BaseModel):
    type: MessageComponentTypes

    def __str__(self):
        """
        for plain text extraction
        :return: human readable text component
        """
        return ''
